#version 400

uniform mat4 modelViewProjectionMatrix;
uniform mat4 inverseModelViewProjectionMatrix;

uniform sampler3D volumeTexture;

uniform float transferFunctionMidValue;
uniform float transferFunctionValueRange;
uniform float marchDistance;
uniform float densityMultiplier;
uniform vec3 tfHue;
uniform int mode;

in vec2 fragPosition;
out vec4 fragColor;

float tfInterpolateBezierCubic(float valueToMidLength, float totalLength)
{
    float t = 1.0 - (valueToMidLength / totalLength);
    return pow(t, 3.0) - (3.0 * pow(t, 2.0)) / 2.0 + (3.0 * t) / 2.0;
}

vec4 transferFunction(float dataValue)
{
    /* [0, 1] */
    float sizeofUint = 65536.0;

    float midValue = transferFunctionMidValue;
    midValue = midValue / sizeofUint;

    float valueRange = transferFunctionValueRange;
    valueRange = valueRange / sizeofUint;

    float valueToMidLength = abs(dataValue - midValue);
    float totalRangeLength = valueRange / 2.0;

    if (valueToMidLength > totalRangeLength)
    {
        return vec4(0.0, 0.0, 0.0, 0.0);
    }

    else
    {
        float totalInfluence = tfInterpolateBezierCubic(valueToMidLength, totalRangeLength);
        totalInfluence *= densityMultiplier;

        vec3 returnColor = vec3(totalInfluence, totalInfluence, totalInfluence);
        float returnAlpha = totalInfluence;

        return vec4(returnColor, returnAlpha);
    }
}

vec4 dvr(vec3 rayOrigin, vec3 rayDirection, vec3 enterPoint, vec3 exitPoint)
{
    float maxDistance = length(exitPoint - enterPoint);
    vec3 currentPosition;
    float totalDistanceTraveled = 0.0;

    vec3 currentColor = vec3(0.0);
    float currentAlpha = 0.0;

    vec3 accumulatedColor = vec3(0.0);
    float accumulatedAlpha = 0.0;

    // Ray marching loop
    while (totalDistanceTraveled < maxDistance)
    {
        currentPosition = enterPoint + totalDistanceTraveled * rayDirection;
        float volumeTextureValue = texture(volumeTexture, currentPosition).r;

        // Map radiodensity to RGBa tuple
        vec4 currentSample = transferFunction(volumeTextureValue);
        currentColor = currentSample.xyz;
        currentAlpha = currentSample.w;

        // Compositing
        float colorAlpha = currentAlpha - (currentAlpha * accumulatedAlpha);
        colorAlpha *= densityMultiplier;
        accumulatedAlpha += colorAlpha;
        accumulatedColor += currentColor * colorAlpha;

        // Early ray-termination
        if (accumulatedAlpha > 0.99)
        {
            break;
        }

        // March along
        totalDistanceTraveled += marchDistance;
    }

    return vec4(tfHue * accumulatedColor, accumulatedAlpha);
}

vec4 mpi(vec3 rayOrigin, vec3 rayDirection, vec3 enterPoint, vec3 exitPoint)
{
    float maxDistance = length(exitPoint - enterPoint);
    vec3 currentPosition;
    float totalDistanceTraveled = 0.0;
    float maximumIntensity = 0.0;

    // Ray marching loop
    while (totalDistanceTraveled < maxDistance)
    {
        vec3 currentPosition = enterPoint + totalDistanceTraveled * rayDirection;
        float densityValue = texture(volumeTexture, currentPosition).r;

        if (densityValue > maximumIntensity)
            maximumIntensity = densityValue;

        // Early ray-termination
        if (maximumIntensity > 0.99)
        {
            break;
        }

        // March along
        totalDistanceTraveled += marchDistance;
    }

    return vec4(tfHue * densityMultiplier * 25 * vec3(maximumIntensity), 1.0);
}

vec4 aip(vec3 rayOrigin, vec3 rayDirection, vec3 enterPoint, vec3 exitPoint)
{
    float maxDistance = length(exitPoint - enterPoint);
    vec3 currentPosition;
    float totalDistanceTraveled = 0.0;
    float averageIntensity = 0.0;

    // Ray marching loop
    while (totalDistanceTraveled < maxDistance)
    {
        vec3 currentPosition = enterPoint + totalDistanceTraveled * rayDirection;
        float densityValue = texture(volumeTexture, currentPosition).r;

        averageIntensity += densityValue;

        // March along
        totalDistanceTraveled += marchDistance;
    }

    return vec4(tfHue * densityMultiplier * 0.8 * vec3(averageIntensity), 1.0);
}

void main()
{
    // Ray setup
    vec4 near = inverseModelViewProjectionMatrix * vec4(fragPosition, -1.0, 1.0);
    near /= near.w;

    vec4 far = inverseModelViewProjectionMatrix * vec4(fragPosition, 1.0, 1.0);
    far /= far.w;

    vec3 rayOrigin = near.xyz;
    vec3 rayDirection = normalize((far - near).xyz);

    // Naive ray segment
    vec3 enterPoint = rayOrigin;
    vec3 exitPoint = rayOrigin + (rayDirection * 5);

    if (mode == 0)
    {
        fragColor = dvr(rayOrigin, rayDirection, enterPoint, exitPoint);
    }
    if (mode == 1)
    {
        fragColor = mpi(rayOrigin, rayDirection, enterPoint, exitPoint);
    }
    if (mode == 2)
    {
        fragColor = aip(rayOrigin, rayDirection, enterPoint, exitPoint);
    }
}
